import numpy as np
import scipy.sparse as sparse
from sklearn.preprocessing import normalize

try:
    from mosek.fusion import *
except:
    print("Warning: Could not load module named mosek.fusion")

try:
    import cvxpy as cvx
except:
    print("Warning: Could not load module named cvxpy")

def log_mvn_pdf(X, atoms, prec, prec_type, homoscedastic):
    """
    given:
       an n x d observation ndarray X,
       an m x d ndarray of atoms,
       an ndarray of precision matrices, whose shape 
           ndarray of shape 
                    (n, d, d) if heteroscedastic & prec_type==general  
                    (n, d)    if heteroscedastic & prec_type==diagonal
                    (d, d)    if homoscedastic & prec_type==general
                    (d,)      if homoscedastic & prec_type==diagonal
    return:
       an n x m matrix of log-probabilities A, where
       A[i, j] = log MVN(X[i] | atoms[j], prec[i]),
       and MVN(X | μ, Ω) is the density of a multivariate 
       normal with mean μ and precision Ω evaluated at X.
    """
    n, d = X.shape
    
    ## the (squared) mahalanobis distance M between all pairs
    ## and the log-determinant ld of the precision matrices
    if prec_type=='general' and not homoscedastic:
        diffs = (X[:, np.newaxis, :] - atoms[np.newaxis, :, :])
        M = np.einsum('ijk,ikl,ijl->ij',diffs,prec,diffs)
        ld = np.linalg.slogdet(prec)[1][:, np.newaxis]
    elif prec_type=='diagonal' and not homoscedastic:
        diffs = (X[:, np.newaxis, :] - atoms[np.newaxis, :, :])
        M = np.einsum('ik,ijk->ij',prec,diffs**2)
        ld = np.sum(np.log(prec), 1)[:, np.newaxis]
    elif prec_type=='general' and homoscedastic:
        M = (np.einsum('ik,kl,il->i',X,prec,X)[:, np.newaxis]
             + np.einsum('jk,kl,jl->j',atoms,prec,atoms)[np.newaxis, :]
             - 2*np.einsum('ik,kl,jl->ij',X,prec,atoms))
        ld = np.linalg.slogdet(prec)[1]
    elif prec_type=='diagonal' and homoscedastic:
        M = ((X**2 @ prec)[:, np.newaxis] 
             + (atoms**2 @ prec)[np.newaxis, :] 
             - 2*np.einsum('ik,k,jk->ij',X,prec,atoms))
        ld = np.sum(np.log(prec))
    else:
        raise Exception("prec_type must be 'general' or 'diagonal'")
    return(-(M + d*np.log(2*np.pi) - ld) / 2)

def mvn_pdf(X, atoms, prec, prec_type, homoscedastic, 
            n_chunks=1, log_prob_thresh=-float('Inf'), row_condition=False):
    """
    given:
       an n x d observation ndarray X,
       an m x d ndarray of atoms,
       an ndarray of precision matrices, of shape 
                    (n, d, d) if heteroscedastic & prec_type==general  
                    (n, d)    if heteroscedastic & prec_type==diagonal
                    (d, d)    if homoscedastic & prec_type==general
                    (d,)      if homoscedastic & prec_type==diagonal
    additional options:
       n_chunks, int default 1
           compute the kernel in chunks
       log_prob_thresh, float default -∞
           set all log probabilities under log_prob_thresh
           to -∞ (so the resulting probability is exactly zero)
       row_condition, bool default False
           if True, divide each probability by the max in each row,
           so the maximum entry in each row is guaranteed to be one
    return:
       an n x m sparse array (csr_matrix) of probabilities K, where
       K[i, j] = MVN(X[i] | atoms[j], prec[i]),
       and MVN(X | μ, Ω) is the density of a multivariate 
       normal with mean μ and precision Ω evaluated at X.
    """
    n = X.shape[0]
    ind_list = np.array_split(np.arange(n), n_chunks)
    res = []
    for inds in ind_list:
        A = log_mvn_pdf(X[inds], atoms, 
                        prec if homoscedastic else prec[inds], 
                        prec_type, homoscedastic)
        if row_condition:
            A = A - np.max(A, 1)[:, np.newaxis]
        A[A < log_prob_thresh] = -float('Inf')
        res.append(sparse.csr_matrix(np.exp(A)))
    return(sparse.vstack(res))


def solve_weights_cvx(K, **solver_params):
    """
    given:
       an n x m kernel K csr_matrix of 
       probabilities, scaled by row
    return:
       the weights w_1,…,w_m maximizing 
        sum_{i=1}^n log(Aw)_i    
       using cvxpy
    """
    n,m = K.shape
    A = K.toarray()

    w = cvx.Variable(m)
    constraints = [w >= 0, cvx.sum(w) == 1]
    obj = cvx.Maximize(cvx.sum(cvx.log(A @ w)))

    prob = cvx.Problem(obj, constraints)
    prob.solve(**solver_params)

    return(w.value)

def solve_weights_mosek(K, use_sparse=True, **solver_params):
    """
    given:
       an n x m kernel K csr_matrix of 
       probabilities, scaled by row
    return:
       the weights w_1,…,w_m maximizing 
        sum_{i=1}^n log(Aw)_i    
       via an exponential cone program in mosek
    """
    n,m = K.shape
    M = Model()

    t = M.variable(n)
    u = M.variable(n, Domain.greaterThan(0.0))
    w = M.variable(m, Domain.inRange(0.0, 1.0))

    ## exponential cone constraints
    for i in range(n):
        ## constrains t[i] ≤ log(u[i]) and u[i] ≥ 0
        M.constraint(Expr.hstack(u.index(i), 1, t.index(i)), 
                     Domain.inPExpCone())
    M.constraint(Expr.sum(w), Domain.equalsTo(1.0))
    if use_sparse:
        K = K.tocoo()
        A = Matrix.sparse(K.row, K.col, K.data)
    else:
        A = Matrix.sparse(K.todense())
    M.constraint(Expr.sub(Expr.mul(A, w), u), Domain.equalsTo(0.0))

    ## Set the objective function to sum(t)
    M.objective("obj", ObjectiveSense.Maximize, Expr.sum(t))
    
    for k,v in solver_params.items():
        M.setSolverParam(k, v)
    
    ## Solve
    M.solve()
    
    return w.level()

class GLMixture:
    """
    
    Kiefer-Wolfowitz nonparametric maximum likelihood estimation 
    (NPMLE) for Gaussian location mixtures in general dimensions,
    supporting general heteroscedastic errors. 


    ----------------------------OPTIONS-----------------------------
    
    prec_type : string, default 'general'
        Type of covariance matrix. Either 'general' or 'diagonal'.
            
    homoscedastic : bool, default False
    
    atoms_init : array, default None
        Specify a set of atoms to use in the discretization.
    
    --------------------------ATTRIBUTES----------------------------
    
    d : int (ambient dimension)
    
    m : int (number of atoms)
    
    n : int (number of training obs)
    
    weights : ndarray of shape (m,)
    
    atoms : ndarray of shape (m, d)
    
    XTrain : ndarray of shape (n, d)
    
    PrecTrain : ndarray of shape 
                    (n, d, d) if heteroscedastic & general  
                    (n, d)    if heteroscedastic & diagonal
                    (d, d)    if homoscedastic & general
                    (d,)      if homoscedastic & diagonal

    """
    
    def __init__(self, prec_type='general', 
                 homoscedastic=False, atoms_init=None):
        self.prec_type = prec_type
        self.homoscedastic = homoscedastic
        self.atoms_init = atoms_init
        
    def get_params(self):
        return(self.atoms, self.weights)
    
    def set_params(self, atoms, weights):
        self.atoms, self.weights = atoms, weights
        
    def initialize_atoms_subsample(self, X, n_atoms):
        self.atoms_init = X[np.random.choice(X.shape[0], 
                                             size=n_atoms, 
                                             replace=False)]
    
    def fit(self, X, prec, max_iter_em=10, weight_thresh=0., 
            n_chunks=1, log_prob_thresh=-float('Inf'), row_condition=False, 
            score_every=1, solver='cvxpy', **solver_params):
        """
        given:
           an n x d observation ndarray X,
           an ndarray of precision matrices, of shape 
                        (n, d, d) if heteroscedastic & prec_type==general  
                        (n, d)    if heteroscedastic & prec_type==diagonal
                        (d, d)    if homoscedastic & prec_type==general
                        (d,)      if homoscedastic & prec_type==diagonal
           max_iter_em : int, default 10
               number of iterations of EM
           weight_thresh : float, default 0
               threshold value for discretized NPMLE weights
           n_chunks : int, default 1
               compute the kernel in chunks
           log_prob_thresh : float, default -∞
               set all log probabilities under log_prob_thresh
               to -∞ (so the resulting probability is exactly zero)
           row_condition : bool, default False
               if True, divide each probability by the max in each row,
               so the maximum entry in each row is guaranteed to be one
           score_every : int, default 1
               number of EM iterations between computing avg log-likelihood
               if None, `fit` method will not call the `score` method
           solver_params : 
               additional parameters for mosek conic interior solver

        
        Approximate the NPMLE by 
           (1) computing the discretized NPMLE over a set atoms_init
           (2) running EM to improve atoms and weights jointly
        Updates self.atoms and self.weights
         
        return:
           a list of average log-likelihoods across EM iterations
        """
        self.n, self.d = X.shape
        self.XTrain = X
        self.PrecTrain = prec
        
        ## initialize atoms
        if self.atoms_init is None:
            print('Selecting all data points as atoms: done.')
            self.atoms_init = X
        
        ## compute sparse kernel matrix
        print('Computing kernel matrix:', end=' ')
        K = mvn_pdf(X, self.atoms_init, prec, self.prec_type, 
                    self.homoscedastic, n_chunks, 
                    log_prob_thresh, row_condition)
        print('done.')
        
        ## solve for the optimal weights given locations atoms_init
        print('Solving for discretized NPMLE:', end=' ')
        if solver=='mosek':
            w = solve_weights_mosek(K, 
                                   use_sparse=(log_prob_thresh > -float('Inf')), 
                                   **solver_params)
        elif solver=='cvxpy':
            w = solve_weights_cvx(K, # no use_sparse option
                                  **solver_params)
        print('done.')
        
        ## threshold weights
        atoms = self.atoms_init[w > weight_thresh]
        weights = w[w > weight_thresh]
        weights /= sum(weights)
        self.set_params(atoms, weights)
        if score_every is not None: llik = [self.score(X, prec)]
        
        if self.prec_type=='diagonal' and not self.homoscedastic:
            precX = np.multiply(prec, X)
        
        ## optimize over weights and locations using EM 
        for _ in range(max_iter_em):
            ## E-step : compute responsibilities
            B = mvn_pdf(X, atoms, prec, self.prec_type, 
                        self.homoscedastic, n_chunks, 
                        log_prob_thresh, row_condition=False)
            resp = B.multiply(weights)
            resp = normalize(resp, norm='l1', axis=1, copy=False)

            ## M-step : update cluster centers
            Nk = np.array(resp.sum(0).T)
            weights = np.ravel(Nk / np.sum(Nk))
            if self.homoscedastic:
                atoms = np.array((resp.transpose().dot(X)) / Nk)
            else:
                if self.prec_type=='general':
                    resp = np.array(resp.todense())
                    Z = np.einsum('ij,ikl->jkl', resp, prec)
                    N = np.einsum('ij,ikl,il->jk', resp, prec, X)
                    for j in range(atoms.shape[0]):
                        atoms[j] = np.linalg.inv(Z[j]) @ N[j]
                elif self.prec_type=='diagonal':
                    num = resp.transpose().dot(precX)
                    dnm = resp.transpose().dot(prec)
                    atoms = np.array(num/dnm)
                    
            if (score_every is not None) and (_ % score_every == 0): 
                print('Running EM: iteration %s / %s' %(_, max_iter_em), 
                      end='\r', flush=True)
                self.set_params(atoms, weights)
                llik.append(self.score(X, prec))
       
        if (score_every is not None): 
            print(' ' * (30 + 2*len(str(max_iter_em))), end='\r', flush=True)    
            print('Running EM: done.', end='\r', flush=True)
            return llik
        
    def score(self, X, prec, n_chunks=1, 
              log_prob_thresh=-float('Inf')):
        """
        given:
           an n x d observation matrix X,
           an array of precision matrices, whose shape depends
           on prec_type and homoscedastic
        return:
           the average log-likelihood across observations X_i.
        """
        K = mvn_pdf(X, self.atoms, prec, self.prec_type, 
                    self.homoscedastic, n_chunks, 
                    log_prob_thresh, row_condition=False)
        return(np.mean(np.log(K.dot(self.weights))))
        
    def posterior_mean(self, X, prec, row_condition=True):
        """
        given:
           an n x d observation matrix X,
           an array of precision matrices, whose shape depends
           on prec_type and homoscedastic
        return:
           the Bayes estimate (posterior mean) for each X_i in 
           the (multivariate, heteroscedastic) normal means
           model, using the prior defined by atoms and weights. 
       """
        a, w = self.get_params()
        A = mvn_pdf(X, a, prec, self.prec_type, self.homoscedastic, 
                row_condition=row_condition)
        return(((A.dot(a * w[:, np.newaxis])) 
               / np.ravel(A.dot(w))[:, np.newaxis]))



import numpy as np
from scipy.interpolate import interp1d
from sklearn.isotonic import isotonic_regression
from sklearn.exceptions import NotFittedError

def check_is_fitted(model):
    if not model.__is_fitted__: raise NotFittedError 

class Grenander:
    """Grenander estimator of a monotone decreasing density.

    Parameters
    ----------

    x_min : float, default=0
        Lower bound on the domain. 
    
    x_max : float or None, default=float('Inf')
        Upper bound on the domain. If None, the upper bound 
        on the domain is the max of the data supplied to `fit`.

    Attributes
    ----------

    cdf : function
        accepts data and returns the estimated distribution function

    pdf : function
        accepts data and returns the estimated probability density function

    knots : array-like of shape (k,)
        knots of the least concave majorant of the empirical cdf

    heights : array-like of shape (k,)
        heights of the least concave majorant of the empirical cdf
        corresponding to the knots
        
    slopes : array-like of shape (k-1,)
        left-hand slopes of the least concave majorant of the empirical cdf
        corresponding to knots[1:]

    Examples
    ________
    Compute the Grenander estimator

    >>> from npeb import Grenander ### SHOULD CHANGE TO sklearn.isotonic and GrenanderDensity
    >>> x = np.random.beta(.4, 1, 50)
    >>> gren = Grenander(x_max=1).fit(x)
    >>> 
    """
    def __init__(self, x_min=0, x_max=float('Inf')):
        self.x_min = x_min
        self.x_max = x_max
        self.__is_fitted__ = False

    def fit(self, x):
        """Compute the Grenander estimator using isotonic regression.

        Parameters
        ----------

        x : array-like of shape (n,)
            observed data

        Returns
        -------

        fitted_values : array-like of shape (n,)
            the value of the Grenander estimator at x
        
        """
        n = len(x)
        
        I = x.argsort()
        x_= x[I]

        # domain of empirical cdf of x
        W = np.hstack([self.x_min, x_])

        # compute slopes of LCM of empirical cdf
        # NB: these are left-hand slopes, so no slope is defined at x_min
        w = W[1:] - W[:-1]
        slopes = isotonic_regression(np.ones(n)/(n*w), 
                                     sample_weight=w.copy(), 
                                     increasing=False)
        
        # save only the knots for faster computation
        keep = np.ones(n, dtype=bool)
        keep[:-1] = ~np.isclose(slopes[:-1], slopes[1:])

        # compute LCM of empirical cdf
        slopes_ = slopes[keep]
        knots_  = np.hstack([self.x_min, x_[keep]])
        heights_= np.hstack([0, (np.where(keep)[0]+1)/n])

        x_max = self.x_max
        if x_max is not None:
            slopes_ = np.hstack([slopes_, 0])
            knots_ = np.hstack([knots_, x_max])
            heights_ = np.hstack([heights_, 1])

        self.slopes = slopes_
        self.knots = knots_
        self.heights = heights_

        self.cdf = interp1d(self.knots, 
                            self.heights, 
                            kind='linear',
                            assume_sorted=True)
        self.pdf = interp1d(self.knots, 
                            np.hstack([np.nan, self.slopes]), 
                            kind='next',
                            assume_sorted=True)
        
        self.__is_fitted__ = True

        # return fitted values in original order
        fitted_values = slopes[I.argsort()]
        return fitted_values
    
    def get_params(self):
        check_is_fitted(self)
        params = {'knots':self.knots,
                  'heights':self.heights,
                  'slopes':self.slopes,
                  'cdf':self.cdf,
                  'pdf':self.pdf,}
        return params

    def score_samples(self, x):
        check_is_fitted(self)
        return np.log(self.pdf(x))

    def score(self, x):
        return np.sum(self.score_samples(x))

def local_fdr(p, zeta=None): 
    """Estimate the local false discovery rate using the Grenander estimator.

    Parameters
    ----------

    p : array-like of shape (m,)
        observed p-values

    zeta : float or None
        threshold used for estimating null proportion pi0
        if None, the estimate pi0=1 is used

    Returns
    -------

    l : array-like of shape (m,)
        the estimated l-values
    
    fdr : function
        maps a p-value to its estimated lfdr

    """
    if zeta is not None:
        pi0_est = (np.sum(p>zeta)+1)/((1-zeta)*len(p))
    else:
        pi0_est = 1

    gren = Grenander(x_max=1)
    fhat = gren.fit(p)

    return(pi0_est/fhat, lambda t: pi0_est/gren.pdf(t))

def support_line(p, tol):
    """Control the local false discovery rate using the SL procedure.

    Parameters
    ----------

    p : array-like of shape (m,)
        observed p-values

    Returns
    -------

    R : integer
        the number of rejections
    
    tau : float
        the rejection threshold

    rejected : array-like of shape (m,)
        rejected[i] indicates whether p[i] is rejected

    """
    m = len(p)
    p_ = np.sort(p)
    p__ = np.hstack([0, p_])

    T = np.arange(m+1)/m
    R = np.argmin(p__-tol*T)
    tau = p__[R]
    return(R, tau, p <= tau)

if __name__=="__main__":
    import matplotlib.pyplot as plt
    np.random.seed(1234567891)

    n = 50
    x = np.random.beta(.4, 1, n)
    gren = Grenander(x_max=1)

    slopes = gren.fit(x)

    plt.figure(figsize=(8,8))
    plt.plot(np.sort(x), np.arange(1,n+1)/n, 'b.', label='Empirical cdf')
    xx = np.linspace(0, 1, 10000)
    plt.plot(xx, gren.cdf(xx), 'r-', label='Least concave majorant')
    plt.legend()

    ax = plt.gca()
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_position('zero')
    ax.spines['left'].set_position('zero')

    plt.ylim([0, 1.01])
    plt.xlim([0, 1])
    plt.show()

    plt.plot(x, slopes, 'b.')
    plt.step(xx, gren.pdf(xx), color='b')
    plt.title('Density estimate')

    plt.ylim([-.01, 2])
    plt.xlim([0, 1])

    ax = plt.gca()
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_position('zero')
    ax.spines['left'].set_position('zero')

    plt.show()

    n = 10000
    pi0 = .75
    p = np.random.rand(n)
    a, b = .3, 1
    p[7500:] = np.random.beta(a, b, 2500)

    _, lfdr = local_fdr(p, zeta=.8)

    from scipy import stats

    tt = np.linspace(0, np.max(p), 10000)
    plt.plot(tt, pi0/(pi0+(1-pi0)*stats.beta.pdf(tt, a, b)), 
             'k-', label='True lfdr')
    plt.plot(tt, lfdr(tt), 'b-', label='Estimated lfdr')
    plt.legend()
    plt.title('Local false discovery rate')

    plt.ylim([-.01, .31])
    plt.xlim([0, .01])

    ax = plt.gca()
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_position('zero')
    ax.spines['left'].set_position('zero')

    plt.show()



import numpy as np
import pandas as pd
from scipy.stats import poisson

try:
    from mosek.fusion import *
except:
    print("Warning: Could not load module named mosek.fusion")

def pois_pmf(X, atoms):
    if len(X.shape) > 1:
        X = np.concatenate(X)
    if len(atoms.shape) > 1:
        atoms = np.concatenate(atoms)
    X = X[:, np.newaxis]
    atoms = atoms[np.newaxis, :]
    return(poisson.pmf(X, atoms))

# t <= log(x), x>=0
def log(M, t, x):
    M.constraint(Expr.hstack(x, 1, t), Domain.inPExpCone())

def solve_weights_mosek(A, nks, **solver_params):
    """
    given:
       an n x m kernel A of probabilities
       a length n vector nks of counts
    return:
       the weights w_1,...,w_m maximizing 
       sum_{i=1}^n nks_i log(Aw)_i
       via an exponential cone program
    """
    n,m = A.shape
    M = Model()

    t = M.variable(n)
    u = M.variable(n, Domain.greaterThan(0.0))
    w = M.variable(m, Domain.inRange(0.0, 1.0))

    # exponential cone constraints
    for i in range(n):
        log(M, t.index(i), u.index(i))
    M.constraint(Expr.sum(w), Domain.equalsTo(1.0))
    for i in range(n):
        M.constraint(Expr.sub(Expr.dot(A[i], w), u.index(i)), 
                     Domain.equalsTo(0.0))

    # Set the objective function to sum_i t_i * nks_i
    M.objective("obj", ObjectiveSense.Maximize, Expr.dot(t, nks))

    # Solve
    M.solve()
    return w.level()

class PMixture:
    """
    Kiefer-Wolfowitz nonparametric maximum likelihood estimation
    (NPMLE) for Poisson mixtures.

    ----------------------------OPTIONS-----------------------------

    atoms_init : array, default None
        Specify a set of atoms to use in the discretization
    
    --------------------------ATTRIBUTES----------------------------

    m : int (number of atoms)

    n : int (number of training obs)

    weights : ndarray of shape (m,)

    atoms : ndarray of shape (m,)

    XTrain : ndarray of shape (n,)
    """

    def __init__(self, atoms_init=None):
        self.atoms_init = atoms_init
        
    def get_params(self):
        return(self.atoms, self.weights)

    def set_params(self, atoms, weights):
        self.atoms, self.weights = atoms, weights

    def initialize_atoms_subsample(self, X, n_atoms):
        self.atoms_init = X[np.random.choice(X.shape[0], 
                                             size=n_atoms, 
                                             replace=False)]

    def initialize_atoms_grid(self, X, n_atoms):
        self.atoms_init = np.linspace(0, X.max()+1, n_atoms)

    def fit(self, X, **solver_params):
        self.n = len(X)
        self.XTrain = X

        if self.atoms_init is None:
            self.initialize_atoms_grid(X, max(50, int(self.n**.5)))
        atoms = self.atoms_init
        
        value_cts = pd.value_counts(X).sort_index()
        vals, cts = np.array(value_cts.keys()), list(value_cts.values)
            
        A = pois_pmf(vals, atoms)
        weights = solve_weights_mosek(A, cts, **solver_params)
        self.set_params(atoms, np.maximum(weights, 0))

    def posterior_mean(self, X):
        a, w = self.get_params()
        K = pois_pmf(X, a)
        return((K@(a*w)) / (K@w))



from distutils.core import setup
setup(
  name = 'npeb',
  packages = ['npeb'],
  version = '0.0.2', 
  license='MIT',   
  description = 'Nonparametric Empirical Bayes',
  author = 'Jake A. Soloff',
  author_email = 'jakesoloff@gmail.com',     
  url = 'https://github.com/jake-soloff/NPEB', 
  download_url = 'https://github.com/jake-soloff/NPEB/archive/refs/tags/v_001.tar.gz',
  keywords = ['Empirical Bayes', 'NPMLE', 'Mixture models'], 
  install_requires=[      
          'numpy',
          'scipy',
          'scikit-learn',
          'pandas',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Science/Research',    
    'Intended Audience :: Developers',    
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3', 
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
 ],
)
